/*
 * Copyright (C) 2010-2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "WebProcessProxy.h"

#include "ArgumentCoders.h"
#if ENABLE(NETSCAPE_PLUGIN_API)
#include "Attachment.h"
#endif
#include "Decoder.h"
#if HAVE(CVDISPLAYLINK)
#include "DisplayLinkObserverID.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "GPUProcessConnectionInfo.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "GPUProcessConnectionParameters.h"
#endif
#include "HandleMessage.h"
#include "NetworkProcessConnectionInfo.h"
#include "SessionState.h"
#if PLATFORM(GTK) || PLATFORM(WPE)
#include "UserMessage.h"
#endif
#if ENABLE(WEB_AUTHN)
#include "WebAuthnProcessConnectionInfo.h"
#endif
#include "WebCoreArgumentCoders.h"
#include "WebProcessProxyMessages.h"
#include <WebCore/FrameIdentifier.h>
#include <WebCore/PageIdentifier.h>
#if ENABLE(NETSCAPE_PLUGIN_API)
#include <WebCore/PluginData.h>
#endif
#include <WebCore/PrewarmInformation.h>
#include <WebCore/RegistrableDomain.h>
#include <WebCore/SleepDisablerIdentifier.h>
#if ENABLE(NETSCAPE_PLUGIN_API)
#include <wtf/Optional.h>
#endif
#if ENABLE(NETSCAPE_PLUGIN_API) || PLATFORM(COCOA)
#include <wtf/Vector.h>
#endif
#include <wtf/text/WTFString.h>

namespace Messages {

namespace WebProcessProxy {

void ShouldTerminate::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, bool shouldTerminate)
{
    encoder.get() << shouldTerminate;
    connection.sendSyncReply(WTFMove(encoder));
}

#if ENABLE(NETSCAPE_PLUGIN_API)

void GetPlugins::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<WebCore::PluginInfo>& plugins, const Vector<WebCore::PluginInfo>& applicationPlugins, const Optional<Vector<WebCore::SupportedPluginIdentifier>>& supportedPluginIdentifiers)
{
    encoder.get() << plugins;
    encoder.get() << applicationPlugins;
    encoder.get() << supportedPluginIdentifiers;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(NETSCAPE_PLUGIN_API)

void GetPluginProcessConnection::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const IPC::Attachment& connectionHandle, bool supportsAsynchronousInitialization)
{
    encoder.get() << connectionHandle;
    encoder.get() << supportsAsynchronousInitialization;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

void GetNetworkProcessConnection::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::NetworkProcessConnectionInfo& connectionInfo)
{
    encoder.get() << connectionInfo;
    connection.sendSyncReply(WTFMove(encoder));
}

#if ENABLE(GPU_PROCESS)

void GetGPUProcessConnection::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::GPUProcessConnectionInfo& connectionInfo)
{
    encoder.get() << connectionInfo;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if ENABLE(WEB_AUTHN)

void GetWebAuthnProcessConnection::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::WebAuthnProcessConnectionInfo& connectionInfo)
{
    encoder.get() << connectionInfo;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(GTK) || PLATFORM(WPE)

void SendMessageToWebContextWithReply::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebKit::UserMessage&&)>&& completionHandler)
{
    Optional<WebKit::UserMessage> replyMessage;
    decoder >> replyMessage;
    if (!replyMessage) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*replyMessage));
}

void SendMessageToWebContextWithReply::cancelReply(CompletionHandler<void(WebKit::UserMessage&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebKit::UserMessage>::create());
}

void SendMessageToWebContextWithReply::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::UserMessage& replyMessage)
{
    encoder.get() << replyMessage;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if PLATFORM(MAC)

void IsAXAuthenticated::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, bool authenticated)
{
    encoder.get() << authenticated;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

} // namespace WebProcessProxy

} // namespace Messages

namespace WebKit {

void WebProcessProxy::didReceiveWebProcessProxyMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    auto protectedThis = makeRef(*this);
    if (decoder.messageName() == Messages::WebProcessProxy::UpdateBackForwardItem::name())
        return IPC::handleMessage<Messages::WebProcessProxy::UpdateBackForwardItem>(decoder, this, &WebProcessProxy::updateBackForwardItem);
    if (decoder.messageName() == Messages::WebProcessProxy::DidDestroyFrame::name())
        return IPC::handleMessage<Messages::WebProcessProxy::DidDestroyFrame>(decoder, this, &WebProcessProxy::didDestroyFrame);
    if (decoder.messageName() == Messages::WebProcessProxy::DidDestroyUserGestureToken::name())
        return IPC::handleMessage<Messages::WebProcessProxy::DidDestroyUserGestureToken>(decoder, this, &WebProcessProxy::didDestroyUserGestureToken);
    if (decoder.messageName() == Messages::WebProcessProxy::EnableSuddenTermination::name())
        return IPC::handleMessage<Messages::WebProcessProxy::EnableSuddenTermination>(decoder, this, &WebProcessProxy::enableSuddenTermination);
    if (decoder.messageName() == Messages::WebProcessProxy::DisableSuddenTermination::name())
        return IPC::handleMessage<Messages::WebProcessProxy::DisableSuddenTermination>(decoder, this, &WebProcessProxy::disableSuddenTermination);
    if (decoder.messageName() == Messages::WebProcessProxy::SetIsHoldingLockedFiles::name())
        return IPC::handleMessage<Messages::WebProcessProxy::SetIsHoldingLockedFiles>(decoder, this, &WebProcessProxy::setIsHoldingLockedFiles);
    if (decoder.messageName() == Messages::WebProcessProxy::DidExceedActiveMemoryLimit::name())
        return IPC::handleMessage<Messages::WebProcessProxy::DidExceedActiveMemoryLimit>(decoder, this, &WebProcessProxy::didExceedActiveMemoryLimit);
    if (decoder.messageName() == Messages::WebProcessProxy::DidExceedInactiveMemoryLimit::name())
        return IPC::handleMessage<Messages::WebProcessProxy::DidExceedInactiveMemoryLimit>(decoder, this, &WebProcessProxy::didExceedInactiveMemoryLimit);
    if (decoder.messageName() == Messages::WebProcessProxy::DidExceedCPULimit::name())
        return IPC::handleMessage<Messages::WebProcessProxy::DidExceedCPULimit>(decoder, this, &WebProcessProxy::didExceedCPULimit);
    if (decoder.messageName() == Messages::WebProcessProxy::StopResponsivenessTimer::name())
        return IPC::handleMessage<Messages::WebProcessProxy::StopResponsivenessTimer>(decoder, this, &WebProcessProxy::stopResponsivenessTimer);
    if (decoder.messageName() == Messages::WebProcessProxy::DidReceiveMainThreadPing::name())
        return IPC::handleMessage<Messages::WebProcessProxy::DidReceiveMainThreadPing>(decoder, this, &WebProcessProxy::didReceiveMainThreadPing);
    if (decoder.messageName() == Messages::WebProcessProxy::DidReceiveBackgroundResponsivenessPing::name())
        return IPC::handleMessage<Messages::WebProcessProxy::DidReceiveBackgroundResponsivenessPing>(decoder, this, &WebProcessProxy::didReceiveBackgroundResponsivenessPing);
    if (decoder.messageName() == Messages::WebProcessProxy::MemoryPressureStatusChanged::name())
        return IPC::handleMessage<Messages::WebProcessProxy::MemoryPressureStatusChanged>(decoder, this, &WebProcessProxy::memoryPressureStatusChanged);
    if (decoder.messageName() == Messages::WebProcessProxy::DidExceedInactiveMemoryLimitWhileActive::name())
        return IPC::handleMessage<Messages::WebProcessProxy::DidExceedInactiveMemoryLimitWhileActive>(decoder, this, &WebProcessProxy::didExceedInactiveMemoryLimitWhileActive);
    if (decoder.messageName() == Messages::WebProcessProxy::DidCollectPrewarmInformation::name())
        return IPC::handleMessage<Messages::WebProcessProxy::DidCollectPrewarmInformation>(decoder, this, &WebProcessProxy::didCollectPrewarmInformation);
#if PLATFORM(COCOA)
    if (decoder.messageName() == Messages::WebProcessProxy::CacheMediaMIMETypes::name())
        return IPC::handleMessage<Messages::WebProcessProxy::CacheMediaMIMETypes>(decoder, this, &WebProcessProxy::cacheMediaMIMETypes);
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebProcessProxy::RequestHighPerformanceGPU::name())
        return IPC::handleMessage<Messages::WebProcessProxy::RequestHighPerformanceGPU>(decoder, this, &WebProcessProxy::requestHighPerformanceGPU);
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebProcessProxy::ReleaseHighPerformanceGPU::name())
        return IPC::handleMessage<Messages::WebProcessProxy::ReleaseHighPerformanceGPU>(decoder, this, &WebProcessProxy::releaseHighPerformanceGPU);
#endif
#if HAVE(CVDISPLAYLINK)
    if (decoder.messageName() == Messages::WebProcessProxy::StartDisplayLink::name())
        return IPC::handleMessage<Messages::WebProcessProxy::StartDisplayLink>(decoder, this, &WebProcessProxy::startDisplayLink);
#endif
#if HAVE(CVDISPLAYLINK)
    if (decoder.messageName() == Messages::WebProcessProxy::StopDisplayLink::name())
        return IPC::handleMessage<Messages::WebProcessProxy::StopDisplayLink>(decoder, this, &WebProcessProxy::stopDisplayLink);
#endif
#if HAVE(CVDISPLAYLINK)
    if (decoder.messageName() == Messages::WebProcessProxy::SetDisplayLinkPreferredFramesPerSecond::name())
        return IPC::handleMessage<Messages::WebProcessProxy::SetDisplayLinkPreferredFramesPerSecond>(decoder, this, &WebProcessProxy::setDisplayLinkPreferredFramesPerSecond);
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    if (decoder.messageName() == Messages::WebProcessProxy::SendMessageToWebContext::name())
        return IPC::handleMessage<Messages::WebProcessProxy::SendMessageToWebContext>(decoder, this, &WebProcessProxy::sendMessageToWebContext);
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    if (decoder.messageName() == Messages::WebProcessProxy::SendMessageToWebContextWithReply::name())
        return IPC::handleMessageAsync<Messages::WebProcessProxy::SendMessageToWebContextWithReply>(connection, decoder, this, &WebProcessProxy::sendMessageToWebContextWithReply);
#endif
    if (decoder.messageName() == Messages::WebProcessProxy::DidCreateSleepDisabler::name())
        return IPC::handleMessage<Messages::WebProcessProxy::DidCreateSleepDisabler>(decoder, this, &WebProcessProxy::didCreateSleepDisabler);
    if (decoder.messageName() == Messages::WebProcessProxy::DidDestroySleepDisabler::name())
        return IPC::handleMessage<Messages::WebProcessProxy::DidDestroySleepDisabler>(decoder, this, &WebProcessProxy::didDestroySleepDisabler);
    if (decoder.messageName() == Messages::WebProcessProxy::CreateSpeechRecognitionServer::name())
        return IPC::handleMessage<Messages::WebProcessProxy::CreateSpeechRecognitionServer>(decoder, this, &WebProcessProxy::createSpeechRecognitionServer);
    if (decoder.messageName() == Messages::WebProcessProxy::DestroySpeechRecognitionServer::name())
        return IPC::handleMessage<Messages::WebProcessProxy::DestroySpeechRecognitionServer>(decoder, this, &WebProcessProxy::destroySpeechRecognitionServer);
    if (decoder.messageName() == Messages::WebProcessProxy::SystemBeep::name())
        return IPC::handleMessage<Messages::WebProcessProxy::SystemBeep>(decoder, this, &WebProcessProxy::systemBeep);
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED_WITH_MESSAGE("Unhandled message %s to %" PRIu64, description(decoder.messageName()), decoder.destinationID());
}

bool WebProcessProxy::didReceiveSyncWebProcessProxyMessage(IPC::Connection& connection, IPC::Decoder& decoder, UniqueRef<IPC::Encoder>& replyEncoder)
{
    auto protectedThis = makeRef(*this);
    if (decoder.messageName() == Messages::WebProcessProxy::ShouldTerminate::name())
        return IPC::handleMessageSynchronous<Messages::WebProcessProxy::ShouldTerminate>(connection, decoder, replyEncoder, this, &WebProcessProxy::shouldTerminate);
#if ENABLE(NETSCAPE_PLUGIN_API)
    if (decoder.messageName() == Messages::WebProcessProxy::GetPlugins::name())
        return IPC::handleMessageSynchronous<Messages::WebProcessProxy::GetPlugins>(connection, decoder, replyEncoder, this, &WebProcessProxy::getPlugins);
#endif
#if ENABLE(NETSCAPE_PLUGIN_API)
    if (decoder.messageName() == Messages::WebProcessProxy::GetPluginProcessConnection::name())
        return IPC::handleMessageSynchronous<Messages::WebProcessProxy::GetPluginProcessConnection>(connection, decoder, replyEncoder, this, &WebProcessProxy::getPluginProcessConnection);
#endif
    if (decoder.messageName() == Messages::WebProcessProxy::GetNetworkProcessConnection::name())
        return IPC::handleMessageSynchronous<Messages::WebProcessProxy::GetNetworkProcessConnection>(connection, decoder, replyEncoder, this, &WebProcessProxy::getNetworkProcessConnection);
#if ENABLE(GPU_PROCESS)
    if (decoder.messageName() == Messages::WebProcessProxy::GetGPUProcessConnection::name())
        return IPC::handleMessageSynchronous<Messages::WebProcessProxy::GetGPUProcessConnection>(connection, decoder, replyEncoder, this, &WebProcessProxy::getGPUProcessConnection);
#endif
#if ENABLE(WEB_AUTHN)
    if (decoder.messageName() == Messages::WebProcessProxy::GetWebAuthnProcessConnection::name())
        return IPC::handleMessageSynchronous<Messages::WebProcessProxy::GetWebAuthnProcessConnection>(connection, decoder, replyEncoder, this, &WebProcessProxy::getWebAuthnProcessConnection);
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::WebProcessProxy::IsAXAuthenticated::name())
        return IPC::handleMessageSynchronous<Messages::WebProcessProxy::IsAXAuthenticated>(connection, decoder, replyEncoder, this, &WebProcessProxy::isAXAuthenticated);
#endif
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
    ASSERT_NOT_REACHED_WITH_MESSAGE("Unhandled synchronous message %s to %" PRIu64, description(decoder.messageName()), decoder.destinationID());
    return false;
}

} // namespace WebKit
